/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 1996   Erich Boleyn  <erich@uruk.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "shared.h"

#include "filesys.h"

#include "fat.h"

static int num_clust;
static int mapblock;
#ifndef PC9800
static int data_offset;
#else
static int sectors_per_record;
#endif
static int fat_size;

/* pointer(s) into filesystem info buffer for DOS stuff */
#define BPB     ( FSYS_BUF + 32256 )  /* 512 bytes long */
#define FAT_BUF ( FSYS_BUF + 30208 )  /* 4 sector FAT buffer */
#define FAT_BUF_SIZE	2048

int
fat_mount(void)
{
  int retval = 1;

  if ( (((DRIVE_HARDDISK_P(current_drive) && !ENTIRE_DISK_P(current_partition))
	 || (current_slice != 0))
	&& !PC_SLICE_TYPE_CHECK(current_slice, FAT12)
	&& !PC_SLICE_TYPE_CHECK(current_slice, FAT16_LT32M)
	&& !PC_SLICE_TYPE_CHECK(current_slice, FAT16_GT32M)
	&& !BSD_FS_TYPE_CHECK(current_slice, MSDOS))
#ifndef PC9800
       || !devread(0, 0, SECTOR_SIZE, BPB)
       || FAT_BPB_BYTES_PER_SECTOR(BPB) != SECTOR_SIZE
#else
       || !devread(0, 0, 128, BPB)
       || (FAT_BPB_BYTES_PER_SECTOR(BPB) & (SECTOR_SIZE(buf_geom) - 1)) != 0
       || ((sectors_per_record = TO_NR_SECTORS(buf_geom,
					       FAT_BPB_BYTES_PER_SECTOR(BPB)))
	   != 1 && sectors_per_record != 2)
#endif
       || FAT_BPB_SECT_PER_CLUST(BPB) < 1 || FAT_BPB_SECT_PER_CLUST(BPB) > 64
       || (FAT_BPB_SECT_PER_CLUST(BPB) & (FAT_BPB_SECT_PER_CLUST(BPB) - 1))
       || !( DRIVE_HARDDISK_P(current_drive)
	     || FAT_BPB_FLOPPY_NUM_SECTORS(BPB) ) )
    retval = 0;
  else
    {
      mapblock = -4096;
#ifndef PC9800
      data_offset = FAT_BPB_DATA_OFFSET(BPB);
#else
#if 0
      root_dir_start = FAT_BPB_ROOT_DIR_START(BPB) * sectors_per_record;
      root_dir_bytes = *(unsigned short *)(BPB + 17) * FAT_DIRENTRY_LENGTH;
      root_dir_sectors
	= TO_NR_SECTORS(buf_geom, root_dir_bytes + SECTOR_SIZE(buf_geom) - 1);
#endif
#endif
      num_clust = FAT_BPB_NUM_CLUST(BPB) + 2;
      if (num_clust > FAT_MAX_12BIT_CLUST)
	fat_size = 4;
      else
	fat_size = 3;
    }

  return retval;
}

static int
fat_create_blocklist(int first_fat_entry)
{
#ifdef PC9800
  int root_start, root_length, root_sectors;
#endif

  BLK_CUR_FILEPOS = 0;
  BLK_CUR_BLKNUM = 0;
  BLK_CUR_BLKLIST = BLK_BLKLIST_START;
  block_file = 1;
  filepos = 0;

#ifdef PC9800
  root_start = FAT_BPB_ROOT_DIR_START(BPB) * sectors_per_record;
  root_length = *(unsigned short *)(BPB + 17) * FAT_DIRENTRY_LENGTH;
  root_sectors = TO_NR_SECTORS(buf_geom,
			       root_length + SECTOR_SIZE(buf_geom) - 1);
#endif

  if (first_fat_entry < 0)
    {
      /* root directory */
#ifndef PC9800
      BLK_BLKSTART(BLK_BLKLIST_START) = FAT_BPB_ROOT_DIR_START(BPB);
      fsmax = filemax = SECTOR_SIZE * (BLK_BLKLENGTH(BLK_BLKLIST_START)
				       = FAT_BPB_ROOT_DIR_LENGTH(BPB));
#else
      BLK_BLKSTART(BLK_BLKLIST_START) = root_start;
      fsmax = filemax = root_length;
      BLK_BLKLENGTH(BLK_BLKLIST_START) = root_sectors;
#endif
    }
  else    /* any real directory/file */
    {
#ifndef PC9800
      int blk_cur_blklist = BLK_BLKLIST_START, blk_cur_blknum;
#else
      int blk_cur_blklist = BLK_BLKLIST_START;
      int sect_per_clust = FAT_BPB_SECT_PER_CLUST(BPB) * sectors_per_record;
#endif
      int last_fat_entry, new_mapblock;

      fsmax = 0;

      do
	{
#ifndef PC9800
	  BLK_BLKSTART(blk_cur_blklist)
	    = (first_fat_entry-2) * FAT_BPB_SECT_PER_CLUST(BPB) + data_offset;

	  blk_cur_blknum = 0;
#else
	  BLK_BLKSTART(blk_cur_blklist)
	    = ((first_fat_entry - 2) * sect_per_clust
	       + root_start + root_sectors);

	  BLK_BLKLENGTH(blk_cur_blklist) = 0;
#endif
	  do
	    {
#ifndef PC9800
	      blk_cur_blknum += FAT_BPB_SECT_PER_CLUST(BPB);
#else
	      BLK_BLKLENGTH(blk_cur_blklist) += sect_per_clust;
#endif
	      last_fat_entry = first_fat_entry;

	      /*
	       *  Do FAT table translation here!
	       */

	      new_mapblock = (last_fat_entry * fat_size) >> 1;
	      if (new_mapblock > (mapblock + 2045)
		  || new_mapblock < (mapblock + 3))
		{
#ifndef PC9800
		  mapblock = ( (new_mapblock < 6) ? 0 :
			       ((new_mapblock - 6) & ~0x1FF) );
		  if (!devread((mapblock>>9)+FAT_BPB_FAT_START(BPB),
			       0, SECTOR_SIZE * 4, FAT_BUF))
		    return 0;
#else
		  mapblock = ((new_mapblock < 6) ? 0
			      : ((new_mapblock - 6) & -SECTOR_SIZE(buf_geom)));
		  if (!devread(TO_NR_SECTORS(buf_geom, mapblock)
			       + FAT_BPB_FAT_START(BPB) * sectors_per_record,
			       0, FAT_BUF_SIZE, FAT_BUF))
		    return 0;
#endif
		}

	      first_fat_entry
		= *((unsigned short *) (FAT_BUF + (new_mapblock - mapblock)));

#if 0
	      if (num_clust <= FAT_MAX_12BIT_CLUST)
#else
	      if (fat_size == 3)
#endif
		{
		  if (last_fat_entry & 1)
		    first_fat_entry >>= 4;
		  else
		    first_fat_entry &= 0xFFF;
		}

	      if (first_fat_entry < 2)
		{
		  errnum = ERR_FSYS_CORRUPT;
		  return 0;
		}
	    }
	  while (first_fat_entry == (last_fat_entry + 1)
		 && first_fat_entry < num_clust);

#ifndef PC9800
	  BLK_BLKLENGTH(blk_cur_blklist) = blk_cur_blknum;
	  fsmax += blk_cur_blknum * SECTOR_SIZE;
#else
	  fsmax += NSECTORS_SIZE(buf_geom, BLK_BLKLENGTH(blk_cur_blklist));
#endif
	  blk_cur_blklist += BLK_BLKLIST_INC_VAL;
	}
      while (first_fat_entry < num_clust && blk_cur_blklist < (FAT_BUF - 7));
    }

  return 1;
}


/* XX FAT filesystem uses the block-list filesystem read function,
   so none is defined here.  */


int
fat_dir(char *dirname)
{
  char *rest, ch, filename[13], dir_buf[FAT_DIRENTRY_LENGTH];
  int attrib = FAT_ATTRIB_DIR, map = -1;

/* main loop to find desired directory entry */
loop:

  if (!fat_create_blocklist(map))
    return 0;

  /* if we have a real file (and we're not just printing possibilities),
     then this is where we want to exit */

  if (!*dirname || isspace(*dirname))
    {
      if (attrib & FAT_ATTRIB_DIR)
	{
	  errnum = ERR_BAD_FILETYPE;
	  return 0;
	}

      return 1;
    }

  /* continue with the file/directory name interpretation */

  while (*dirname == '/')
    dirname++;

  filemax = fsmax;

  if (!filemax || !(attrib & FAT_ATTRIB_DIR))
    {
      errnum = ERR_BAD_FILETYPE;
      return 0;
    }

  for (rest = dirname; (ch = *rest) && !isspace(ch) && ch != '/'; rest++) ;

  *rest = 0;

  do
    {
      if (read(dir_buf, FAT_DIRENTRY_LENGTH) != FAT_DIRENTRY_LENGTH)
	{
	  if (!errnum)
	    {
	      if (print_possibilities < 0)
		{
		  putchar('\n');
		  return 1;
		}

	      errnum = ERR_FILE_NOT_FOUND;
	      *rest = ch;
	    }

	  return 0;
	}

      if (!FAT_DIRENTRY_VALID(dir_buf))
	continue;

      /* XXX convert to 8.3 filename format here */
      {
	int i, j, c;

	for (i = 0; i < 8 && (c = filename[i] = tolower(dir_buf[i]))
	       && !isspace(c) ; i++) ;

	filename[i++] = '.';

	for (j = 0; j < 3 && (c = filename[i+j] = tolower(dir_buf[8+j]))
	       && !isspace(c) ; j++) ;

	if (j == 0)
	  i--;

	filename[i+j] = 0;
      }

      if (print_possibilities && ch != '/'
	  && (!*dirname || strcmp(dirname, filename) <= 0))
	{
	  if (print_possibilities > 0)
	    print_possibilities = -print_possibilities;
	  printf("  %s", filename);
	}
    }
  while (strcmp(dirname, filename) != 0 || (print_possibilities && ch != '/'));

  *(dirname = rest) = ch;

  attrib = FAT_DIRENTRY_ATTRIB(dir_buf);
  filemax = FAT_DIRENTRY_FILELENGTH(dir_buf);
  map = FAT_DIRENTRY_FIRST_CLUSTER(dir_buf);

  /* go back to main loop at top of function */
  goto loop;
}


